# ----------------------------------------------------------------------------
# Copyright (c) 2017-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import pandas as pd
from q2_types.distance_matrix import DistanceMatrix
from q2_types.sample_data import SampleData
from q2_types.metadata import ImmutableMetadata
import qiime2.plugin
from qiime2.plugin import (
    Int, Categorical, MetadataColumn, model, Numeric, Plugin, SemanticType,
    Str, Bool, Metadata, ValidationError,
)

from . import tabulate, distance_matrix, shuffle_groups, merge, __version__

plugin = Plugin(
    name='metadata',
    version=__version__,
    website='https://github.com/qiime2/q2-metadata',
    package='q2_metadata',
    user_support_text=None,
    citation_text=None,
    description=('This QIIME 2 plugin provides functionality for working with '
                 'and visualizing Metadata.'),
    short_description='Plugin for working with Metadata.'
)

plugin.methods.register_function(
    function=distance_matrix,
    inputs={},
    parameters={'metadata': MetadataColumn[Numeric]},
    parameter_descriptions={'metadata': 'Numeric metadata column to compute '
                                        'pairwise Euclidean distances from'},
    outputs=[('distance_matrix', DistanceMatrix)],
    name='Create a distance matrix from a numeric Metadata column',
    description='Create a distance matrix from a numeric metadata column. '
                'The Euclidean distance is computed between each pair of '
                'samples or features in the column.\n\n'
                'Tip: the distance matrix produced by this method can be used '
                'as input to the Mantel test available in `q2-diversity`.'
)

plugin.visualizers.register_function(
    function=tabulate,
    inputs={},
    parameters={
        'input': qiime2.plugin.Metadata,
        'page_size': qiime2.plugin.Int,
    },
    parameter_descriptions={
        'input': 'The metadata to tabulate.',
        'page_size': 'The maximum number of Metadata records to display '
                     'per page',
    },
    name='Interactively explore Metadata in an HTML table',
    description='Generate a tabular view of Metadata. The output '
                'visualization supports interactive filtering, sorting, and '
                'exporting to common file formats.',
)

ArtificialGrouping = \
    SemanticType('ArtificialGrouping', variant_of=SampleData.field['type'])

plugin.register_semantic_types(ArtificialGrouping)


class ArtificialGroupingFormat(model.TextFileFormat):
    def validate(self, *args):
        try:
            md = qiime2.Metadata.load(str(self))
        except qiime2.metadata.MetadataFileError as md_exc:
            raise ValidationError(md_exc) from md_exc

        if md.column_count == 0:
            raise ValidationError('Format must contain at least 1 column')

        filtered_md = md.filter_columns(column_type='categorical')
        if filtered_md.column_count != md.column_count:
            raise ValidationError('Must only contain categorical values.')


ArtificialGroupingDirectoryFormat = model.SingleFileDirectoryFormat(
    'ArtificialGroupingDirectoryFormat', 'artificial-groupings.tsv',
    ArtificialGroupingFormat)

plugin.register_formats(ArtificialGroupingFormat,
                        ArtificialGroupingDirectoryFormat)

plugin.register_semantic_type_to_format(
    SampleData[ArtificialGrouping],
    artifact_format=ArtificialGroupingDirectoryFormat)


@plugin.register_transformer
def _1(df: pd.DataFrame) -> (ArtificialGroupingFormat):
    ff = ArtificialGroupingFormat()
    md = qiime2.Metadata(df)
    md.save(str(ff))
    return ff


@plugin.register_transformer
def _2(ff: ArtificialGroupingFormat) -> (qiime2.Metadata):
    return qiime2.Metadata.load(str(ff))


plugin.methods.register_function(
    function=shuffle_groups,
    inputs={},
    parameters={'metadata': MetadataColumn[Categorical],
                'n_columns': Int,
                'md_column_name_prefix': Str,
                'md_column_values_prefix': Str,
                'encode_sample_size': Bool
                },
    parameter_descriptions={
        'metadata': ('Categorical metadata column to shuffle.'),
        'n_columns': 'The number of shuffled metadata columns to create.',
        'md_column_name_prefix': ('Prefix to use in naming the shuffled '
                                  'metadata columns.'),
        'md_column_values_prefix': ('Prefix to use in naming the values in '
                                    'the shuffled metadata columns.'),
        'encode_sample_size': ('If true, the sample size of each metadata '
                               'group will be appended to the shuffled '
                               'metadata column values.'),
        },
    output_descriptions={
        'shuffled_groups': 'Randomized metadata columns'},
    outputs=[('shuffled_groups', SampleData[ArtificialGrouping])],
    name='Shuffle values in a categorical sample metadata column.',
    description=('Create one or more categorical sample metadata '
                 'columns by shuffling the values in an input metadata '
                 'column. To avoid confusion, the column name and values '
                 'will be derived from the provided prefixes. The number of '
                 'different values (or groups), and the counts of each value, '
                 'will match the input metadata column but the association of '
                 'values with sample ids will be random. These data will be '
                 'written to an artifact that can be used as sample metadata.')
)


plugin.methods.register_function(
    function=merge,
    inputs={},
    parameters={'metadata1': Metadata,
                'metadata2': Metadata},
    parameter_descriptions={
        'metadata1': 'First metadata file to merge.',
        'metadata2': 'Second metadata file to merge.'
    },
    outputs=[('merged_metadata', ImmutableMetadata)],
    output_descriptions={
        'merged_metadata': 'The merged metadata.'
    },
    name='Merge metadata',
    description=('Merge metadata that contains overlapping ids or overlapping '
                 'columns, but not both overlapping ids and overlapping '
                 'columns. The result will be the union (i.e., outer join) '
                 'of the ids and columns from the two metadata inputs.\n\n'
                 'Attemping to merge metadata with both overlapping ids and '
                 'overlapping columns will currently fail because we don\'t '
                 'resolve conflicting column values for a sample. '
                 'Merging metadata with neither overlapping ids or '
                 'overlapping columns is possible with this action.\n\n'
                 'To merge more than two metadata objects, run this command '
                 'multiple times, iteratively using the output of the '
                 'previous run as one of the metadata inputs.\n\n'
                 'The output, an ImmutableMetadata artifact, can be used '
                 'anywhere that a metadata file can be used, or can be '
                 'exported to a metadata tsv file in the typical format.')
)



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import unittest

import pandas as pd
import qiime2

from q2_metadata import shuffle_groups


class ShuffleGroupsTests(unittest.TestCase):

    # number of iterations to run for tests of randomization
    n_iterations = 500

    def test_shuffle_groups_shape_41(self):
        md = qiime2.CategoricalMetadataColumn(
            pd.Series(['a', 'b', 'a', 'b'], name='groups',
                      index=pd.Index(['sample1', 'sample2', 'sample3', 's4'],
                                     name='id'))
        )

        # expected number of rows and columns in result
        obs = shuffle_groups(md, n_columns=1,
                             md_column_name_prefix='shuffled.grouping.',
                             md_column_values_prefix='fake.group.')
        self.assertEqual(obs.shape, (4, 1))

        # expected column names (the original should not be in the result)
        self.assertFalse('groups' in obs.columns)
        self.assertTrue('shuffled.grouping.0' in obs.columns)

        # correct number of groups in the new column
        self.assertEqual(len(obs['shuffled.grouping.0'].unique()), 2)

        # correct group names in new column
        self.assertEqual(set(obs['shuffled.grouping.0'].unique()),
                         {'fake.group.1', 'fake.group.0'})

        # distributions of value counts are equal in input and output
        self.assertEqual(
            sorted(list(obs['shuffled.grouping.0'].value_counts())),
            sorted(list(md.to_series().value_counts())))

        # randomization of key/value associations is occurring
        random_check = []
        for i in range(self.n_iterations):
            obs2 = shuffle_groups(md, n_columns=1,
                                  md_column_name_prefix='shuffled.grouping.',
                                  md_column_values_prefix='fake.group.')
            random_check.append(
                list(obs['shuffled.grouping.0']) ==
                list(obs2['shuffled.grouping.0']))
        self.assertIn(False, random_check,
                      "All random groupings in %d iterations were "
                      "identicial, suggesting that values are not "
                      "randomly assigned." % self.n_iterations)

    def test_shuffle_groups_shape_33(self):
        md = qiime2.CategoricalMetadataColumn(
            pd.Series(['a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c'],
                      name='groups',
                      index=pd.Index(['sample1', 'sample2', 'sample3',
                                      'samplea', 'sampleb', 'sc',
                                      'sample1_w', 'ctl1', 'ctl3'],
                                     name='id'))
        )

        # expected number of rows and columns
        obs = shuffle_groups(md, n_columns=3,
                             md_column_name_prefix='shuffled.grouping.',
                             md_column_values_prefix='fake.group.')
        self.assertEqual(obs.shape, (9, 3))

        # original column name should not be in the result
        self.assertFalse('groups' in obs.columns)

        for i in range(3):
            column_id = 'shuffled.grouping.%d' % i
            self.assertTrue(column_id in obs.columns)

            # correct number of groups in the new column
            self.assertEqual(len(obs[column_id].unique()), 3)

            self.assertEqual(
                set(obs[column_id].unique()),
                {'fake.group.1', 'fake.group.0', 'fake.group.2'})

        # randomization of key/value associations is occurring
        random_check1 = []
        random_check2 = []
        random_check3 = []
        for i in range(self.n_iterations):
            random_check1.append(
                list(obs['shuffled.grouping.0']) ==
                list(obs['shuffled.grouping.1']))
            random_check2.append(
                list(obs['shuffled.grouping.0']) ==
                list(obs['shuffled.grouping.2']))
            random_check3.append(
                list(obs['shuffled.grouping.1']) ==
                list(obs['shuffled.grouping.2']))
        self.assertIn(
            False, random_check1,
            "All random groupings in %d iterations were "
            "identicial, suggesting that values are not "
            "randomly assigned." % self.n_iterations)
        self.assertIn(
            False, random_check2,
            "All random groupings in %d iterations were "
            "identicial, suggesting that values are not "
            "randomly assigned." % self.n_iterations)
        self.assertIn(
            False, random_check3,
            "All random groupings in %d iterations were "
            "identicial, suggesting that values are not "
            "randomly assigned." % self.n_iterations)

    def test_shuffle_groups_alt_input_column_name(self):
        md = qiime2.CategoricalMetadataColumn(
            pd.Series(['a', 'b', 'a', 'b'], name='xyz',
                      index=pd.Index(['sample1', 'sample2', 'sample3', 's4'],
                                     name='id'))
        )

        # expected number of rows and columns in result
        obs = shuffle_groups(md, n_columns=1,
                             md_column_name_prefix='shuffled.grouping.',
                             md_column_values_prefix='fake.group.')
        self.assertEqual(obs.shape, (4, 1))

        # expected column names (the original should not be in the result)
        self.assertFalse('xyz' in obs.columns)
        self.assertTrue('shuffled.grouping.0' in obs.columns)

        # correct number of groups in the new column
        self.assertEqual(len(obs['shuffled.grouping.0'].unique()), 2)

        # correct group names in new column
        self.assertEqual(set(obs['shuffled.grouping.0'].unique()),
                         {'fake.group.1', 'fake.group.0'})

    def test_shuffle_groups_alt_column_name_prefix(self):
        md = qiime2.CategoricalMetadataColumn(
            pd.Series(['a', 'b', 'a', 'b'], name='groups',
                      index=pd.Index(['sample1', 'sample2', 'sample3', 's4'],
                                     name='id'))
        )

        # expected number of rows and columns in result
        obs = shuffle_groups(md, n_columns=1,
                             md_column_name_prefix='1',
                             md_column_values_prefix='fake.group.')
        self.assertEqual(obs.shape, (4, 1))

        # expected column names (the original should not be in the result)
        self.assertFalse('groups' in obs.columns)
        self.assertTrue('10' in obs.columns)

        # correct number of groups in the new column
        self.assertEqual(len(obs['10'].unique()), 2)

        # correct group names in new column
        self.assertEqual(set(obs['10'].unique()),
                         {'fake.group.1', 'fake.group.0'})

    def test_shuffle_groups_alt_column_value_prefix(self):
        md = qiime2.CategoricalMetadataColumn(
            pd.Series(['a', 'b', 'a', 'b'], name='groups',
                      index=pd.Index(['sample1', 'sample2', 'sample3', 's4'],
                                     name='id'))
        )

        # expected number of rows and columns in result
        obs = shuffle_groups(md, n_columns=1,
                             md_column_name_prefix='shuffled.grouping.',
                             md_column_values_prefix='1')
        self.assertEqual(obs.shape, (4, 1))

        # expected column names (the original should not be in the result)
        self.assertFalse('groups' in obs.columns)
        self.assertTrue('shuffled.grouping.0' in obs.columns)

        # correct number of groups in the new column
        self.assertEqual(len(obs['shuffled.grouping.0'].unique()), 2)

        # correct group names in new column
        self.assertEqual(
            set(obs['shuffled.grouping.0'].unique()),
            {'11', '10'})

    def test_shuffle_groups_sample_size_column_id_flag_no_input(self):
        md = qiime2.CategoricalMetadataColumn(
            pd.Series(['a', 'b', 'a', 'b'], name='groups',
                      index=pd.Index(['sample1', 'sample2', 'sample3', 's4'],
                                     name='id'))
        )

        # expected number of rows and columns in result
        obs = shuffle_groups(md, n_columns=1,
                             md_column_name_prefix='shuffled.grouping.',
                             md_column_values_prefix='fake.group.',
                             )

        # correct group names in new column
        self.assertEqual(
            set(obs['shuffled.grouping.0'].unique()),
            {'fake.group.1', 'fake.group.0'})

    def test_shuffle_groups_sample_size_column_id_flag_true(self):
        md = qiime2.CategoricalMetadataColumn(
            pd.Series(['a', 'b', 'a', 'b'], name='groups',
                      index=pd.Index(['sample1', 'sample2', 'sample3', 's4'],
                                     name='id'))
        )

        # expected number of rows and columns in result
        obs = shuffle_groups(md, n_columns=1,
                             md_column_name_prefix='shuffled.grouping.',
                             md_column_values_prefix='fake.group.',
                             encode_sample_size=True)

        # correct group names in new column
        self.assertEqual(
            set(obs['shuffled.grouping.0'].unique()),
            {'fake.group.1.n=2', 'fake.group.0.n=2'})

    def test_shuffle_groups_sample_size_column_id_flag_false(self):
        md = qiime2.CategoricalMetadataColumn(
            pd.Series(['a', 'b', 'a', 'b'], name='groups',
                      index=pd.Index(['sample1', 'sample2', 'sample3', 's4'],
                                     name='id'))
        )

        # expected number of rows and columns in result
        obs = shuffle_groups(md, n_columns=1,
                             md_column_name_prefix='shuffled.grouping.',
                             md_column_values_prefix='1',
                             encode_sample_size=False)

        # correct group names in new column
        self.assertEqual(
            set(obs['shuffled.grouping.0'].unique()),
            {'11', '10'})

    def test_shuffle_groups_sample_size_equals_value_counts(self):
        md = qiime2.CategoricalMetadataColumn(
            pd.Series(['a', 'b', 'b', 'b', 'a', 'b'], name='groups',
                      index=pd.Index(['sample1', 'sample2', 'sample3', 's4',
                                      's5', 's6'],
                                     name='id'))
        )

        # expected number of rows and columns in result
        obs = shuffle_groups(md, n_columns=1,
                             md_column_name_prefix='shuffled.grouping.',
                             md_column_values_prefix='fake.group.',
                             encode_sample_size=True)

        # expected number of samples in each group in the new column
        self.assertEqual(obs['shuffled.grouping.0'].value_counts()
                         ['fake.group.0.n=2'], 2)
        self.assertEqual(obs['shuffled.grouping.0'].value_counts()
                            ['fake.group.1.n=4'], 4)



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import unittest

import numpy as np
import pandas as pd
import qiime2

from q2_metadata import merge


class MergeTests(unittest.TestCase):

    def test_merge_overlapping_samples_and_columns_errors(self):
        index1 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data1 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]
        md1 = qiime2.Metadata(pd.DataFrame(data1, index=index1, dtype=object,
                                           columns=['col1', 'col2', 'col3']))

        index2 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data2 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]

        md2 = qiime2.Metadata(pd.DataFrame(data2, index=index2, dtype=object,
                                           columns=['col1', 'col2', 'col3']))

        self.assertRaisesRegex(ValueError,
                               "3 overl.*ids.*sample1.*3 overl.*col.*col1",
                               merge, md1, md2)

        index3 = pd.Index(['sample4', 'sample5', 'sample1'], name='id')
        data3 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]
        md3 = qiime2.Metadata(pd.DataFrame(data3, index=index3, dtype=object,
                                           columns=['col4', 'col5', 'col1']))

        self.assertRaisesRegex(ValueError,
                               "1 overl.*ids.*sample1.*1 overl.*col.*col1",
                               merge, md1, md3)

    def test_merge_all_samples_overlapping(self):
        index1 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data1 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]
        md1 = qiime2.Metadata(pd.DataFrame(data1, index=index1, dtype=object,
                                           columns=['col1', 'col2', 'col3']))

        index2 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data2 = [['k', 'n', 'q'],
                 ['l', 'o', 'r'],
                 ['m', 'p', 's']]
        md2 = qiime2.Metadata(pd.DataFrame(data2, index=index2, dtype=object,
                                           columns=['col4', 'col5', 'col6']))

        obs1 = merge(md1, md2)

        index_exp1 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data_exp1 = [['a', 'd', 'h', 'k', 'n', 'q'],
                     ['b', 'e', 'i', 'l', 'o', 'r'],
                     ['c', 'f', 'j', 'm', 'p', 's']]
        exp1 = qiime2.Metadata(
            pd.DataFrame(data_exp1, index=index_exp1, dtype=object,
                         columns=['col1', 'col2', 'col3',
                                  'col4', 'col5', 'col6']))

        self.assertEqual(obs1, exp1)

    def test_merge_some_samples_overlapping(self):
        index1 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data1 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]
        md1 = qiime2.Metadata(pd.DataFrame(data1, index=index1, dtype=object,
                                           columns=['col1', 'col2', 'col3']))

        index2 = pd.Index(['sample1', 'sample2', 'sample4'], name='id')
        data2 = [['k', 'n'],
                 ['l', 'o'],
                 ['m', 'p']]
        md2 = qiime2.Metadata(pd.DataFrame(data2, index=index2, dtype=object,
                                           columns=['col4', 'col5']))

        obs1 = merge(md1, md2)

        index_exp1 = pd.Index(['sample1', 'sample2', 'sample3', 'sample4'],
                              name='id')
        data_exp1 = [['a', 'd', 'h', 'k', 'n'],
                     ['b', 'e', 'i', 'l', 'o'],
                     ['c', 'f', 'j', np.nan, np.nan],
                     [np.nan, np.nan, np.nan, 'm', 'p']]
        exp1 = qiime2.Metadata(
            pd.DataFrame(data_exp1, index=index_exp1, dtype=object,
                         columns=['col1', 'col2', 'col3',
                                  'col4', 'col5']))

        self.assertEqual(obs1, exp1)

    def test_merge_all_columns_overlapping(self):
        index1 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data1 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]
        md1 = qiime2.Metadata(pd.DataFrame(data1, index=index1, dtype=object,
                                           columns=['col1', 'col2', 'col3']))

        index2 = pd.Index(['sample4', 'sample5', 'sample6'], name='id')
        data2 = [['k', 'n', 'q'],
                 ['l', 'o', 'r'],
                 ['m', 'p', 's']]
        md2 = qiime2.Metadata(pd.DataFrame(data2, index=index2, dtype=object,
                                           columns=['col1', 'col2', 'col3']))

        obs1 = merge(md1, md2)
        index_exp1 = pd.Index(['sample1', 'sample2', 'sample3',
                               'sample4', 'sample5', 'sample6'], name='id')
        data_exp1 = [['a', 'd', 'h'],
                     ['b', 'e', 'i'],
                     ['c', 'f', 'j'],
                     ['k', 'n', 'q'],
                     ['l', 'o', 'r'],
                     ['m', 'p', 's']]
        exp1 = qiime2.Metadata(
            pd.DataFrame(data_exp1, index=index_exp1, dtype=object,
                         columns=['col1', 'col2', 'col3']))
        self.assertEqual(obs1, exp1)

    def test_merge_some_columns_overlapping(self):
        index1 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data1 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]
        md1 = qiime2.Metadata(pd.DataFrame(data1, index=index1, dtype=object,
                                           columns=['col1', 'col2', 'col3']))

        index2 = pd.Index(['sample4', 'sample5', 'sample6'], name='id')
        data2 = [['k', 'n', 'q'],
                 ['l', 'o', 'r'],
                 ['m', 'p', 's']]
        md2 = qiime2.Metadata(pd.DataFrame(data2, index=index2, dtype=object,
                                           columns=['col1', 'col2', 'col4']))

        obs1 = merge(md1, md2)

        index_exp1 = pd.Index(['sample1', 'sample2', 'sample3',
                               'sample4', 'sample5', 'sample6'], name='id')
        data_exp1 = [['a', 'd', 'h', np.nan],
                     ['b', 'e', 'i', np.nan],
                     ['c', 'f', 'j', np.nan],
                     ['k', 'n', np.nan, 'q'],
                     ['l', 'o', np.nan, 'r'],
                     ['m', 'p', np.nan, 's']]
        exp1 = qiime2.Metadata(
            pd.DataFrame(data_exp1, index=index_exp1, dtype=object,
                         columns=['col1', 'col2', 'col3', 'col4']))

        self.assertEqual(obs1, exp1)

    def test_merge_no_samples_or_columns_overlapping(self):
        index1 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data1 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]
        md1 = qiime2.Metadata(pd.DataFrame(data1, index=index1, dtype=object,
                                           columns=['col1', 'col2', 'col3']))

        index2 = pd.Index(['sample4', 'sample5', 'sample6'], name='id')
        data2 = [['k', 'n', 'q'],
                 ['l', 'o', 'r'],
                 ['m', 'p', 's']]
        md2 = qiime2.Metadata(pd.DataFrame(data2, index=index2, dtype=object,
                                           columns=['col4', 'col5', 'col6']))

        obs1 = merge(md1, md2)

        index_exp1 = pd.Index(['sample1', 'sample2', 'sample3',
                               'sample4', 'sample5', 'sample6'], name='id')
        data_exp1 = [['a', 'd', 'h', np.nan, np.nan, np.nan],
                     ['b', 'e', 'i', np.nan, np.nan, np.nan],
                     ['c', 'f', 'j', np.nan, np.nan, np.nan],
                     [np.nan, np.nan, np.nan, 'k', 'n', 'q'],
                     [np.nan, np.nan, np.nan, 'l', 'o', 'r'],
                     [np.nan, np.nan, np.nan, 'm', 'p', 's']]
        exp1 = qiime2.Metadata(
            pd.DataFrame(data_exp1, index=index_exp1, dtype=object,
                         columns=['col1', 'col2', 'col3',
                                  'col4', 'col5', 'col6']))

        self.assertEqual(obs1, exp1)

    def test_merge_mismatched_columnID_names_in_error_message(self):
        index1 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data1 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]
        md1 = qiime2.Metadata(pd.DataFrame(data1, index=index1, dtype=object,
                                           columns=['col1', 'col2', 'col3']))

        index2 = pd.Index(['sample4', 'sample5', 'sample6'], name='sample-id')
        data2 = [['k', 'n', 'q'],
                 ['l', 'o', 'r'],
                 ['m', 'p', 's']]
        md2 = qiime2.Metadata(pd.DataFrame(data2, index=index2, dtype=object,
                                           columns=['col4', 'col5', 'col6']))

        with self.assertRaisesRegex(
            ValueError,
            "Metadata files contain different ID column names.*id.*sample-id"
        ):
            merge(md1, md2)

    def test_merge_mismatched_md_column_type_designations(self):
        index1 = pd.Index(['sample1', 'sample2', 'sample3'], name='id')
        data1 = [['a', 'd', 'h'],
                 ['b', 'e', 'i'],
                 ['c', 'f', 'j']]
        md1 = qiime2.Metadata(
            pd.DataFrame(
                data1,
                index=index1,
                dtype=object,
                columns=['col1', 'col2', 'col3']
            )
        )
        index2 = pd.Index(['sample4', 'sample5', 'sample6'], name='id')
        data2 = [['k', 'n', 40.0],
                 ['l', 'o', 41.0],
                 ['m', 'p', 42.0]]
        md2 = qiime2.Metadata(
            pd.DataFrame(
                data2,
                index=index2,
                columns=['col1', 'col2', 'col3']
            )
        )
        with self.assertRaisesRegex(
            ValueError,
            "Metadata files contain the shared column 'col3' with different "
            "type designations. In 'metadata1', the column 'col3' is of type "
            r"\'CategoricalMetadataColumn\', and in 'metadata2', it is of "
            r"type \'NumericMetadataColumn\'. These type designations must "
            "match."
        ):
            merge(md1, md2)



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import qiime2
import numpy as np
import pandas as pd


def shuffle_groups(metadata: qiime2.CategoricalMetadataColumn,
                   n_columns: int = 3,
                   md_column_name_prefix: str = 'shuffled.grouping.',
                   md_column_values_prefix: str = 'fake.group.',
                   encode_sample_size: bool = False
                   ) -> pd.DataFrame:

    input_column_name = metadata.name
    df = metadata.to_dataframe()
    group_sample_size = df[input_column_name].value_counts()

    value_mapping = {}
    for i, value in enumerate(df[input_column_name].unique()):
        if encode_sample_size:
            value_mapping[value] = '%s%d%s' % (md_column_values_prefix, i,
                                               f'.n={group_sample_size[value]}'
                                               )
        else:
            value_mapping[value] = '%s%d' % (md_column_values_prefix, i)

    first_column_id = '%s0' % md_column_name_prefix
    df[first_column_id] = df[input_column_name].map(value_mapping)

    df[first_column_id] = \
        np.random.permutation(df[first_column_id].values)

    for i in range(1, n_columns):
        column_id = '%s%d' % (md_column_name_prefix, i)
        df[column_id] = \
            np.random.permutation(df[first_column_id].values)

    df = df.drop(input_column_name, axis=1)
    return df



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2023, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import qiime2
import pandas as pd


def merge(metadata1: qiime2.Metadata,
          metadata2: qiime2.Metadata) -> qiime2.Metadata:
    # Ultimately it would make sense for this action to take
    # List[qiime2.Metadata] as input, but this isn't possible right now
    overlapping_ids = set(metadata1.ids) & set(metadata2.ids)
    overlapping_columns = set(metadata1.columns) & set(metadata2.columns)
    n_overlapping_ids = len(overlapping_ids)
    n_overlapping_columns = len(overlapping_columns)

    if n_overlapping_ids and n_overlapping_columns:
        raise ValueError(
            "Merging can currently handle overlapping ids or overlapping "
            f"columns but not both. {n_overlapping_ids} overlapping ids were "
            f"identified ({', '.join(overlapping_ids)}) and "
            f"{n_overlapping_columns} overlapping columns were identified "
            f"({', '.join(overlapping_columns)})."
        )

    df1 = metadata1.to_dataframe()
    df2 = metadata2.to_dataframe()

    if df1.index.name != df2.index.name:
        raise ValueError(
            "Metadata files contain different ID column names. "
            f"Metadata1 file contains '{df1.index.name}' and metadata2 "
            f"contains '{df2.index.name}'. These column names must match."
        )

    if not n_overlapping_columns:
        result = pd.merge(df1, df2, how='outer', left_index=True,
                          right_index=True)

    else:
        for column in overlapping_columns:
            if df1[column].dtype != df2[column].dtype:
                column_type1 = type(
                    qiime2.Metadata(df1[[column]]).get_column(column))
                column_type2 = type(
                    qiime2.Metadata(df2[[column]]).get_column(column))
                raise ValueError(
                    f"Metadata files contain the shared column '{column}' "
                    "with different type designations. "
                    f"In 'metadata1', the column '{column}' is of type "
                    f"'{column_type1.__name__}', "
                    f"and in 'metadata2', it is of type "
                    f"'{column_type2.__name__}'. These type designations must "
                    "match."
                )

        result = pd.merge(df1, df2, how='outer', left_index=True,
                          right_index=True, suffixes=('', '_'))
        for c in overlapping_columns:
            result[c] = result[c].combine_first(result[f"{c}_"])
            result = result.drop(columns=[f"{c}_"])

    return qiime2.Metadata(result)
